/*
 * Copyright (c) 2001-2004 Living Systems(R) GmbH, Germany.
 * All rights reserved.
 * Original Author: kdorer
 *
 * $Source: /cvs/Research/lecture/AI/src/domain/IProblemState.java,v $
 * $Date: 2007/01/04 08:20:46 $
 */
package kdo.domain;

import java.util.List;

public interface IProblemState
{
	/**
	 * Checks if this state is a goal state of this problem
	 * @return true if this state is a goal state of this problem
	 */
	public boolean checkGoalState();

	/**
	 * Returns the list of operators applicable in this state. The default
	 * implementation returns a list of operators applicable in all states
	 * @return the list of operators applicable in this state
	 */
	public List<IOperator> getOperators();

	/**
	 * Returns the problem (domain) to which this state belongs
	 * @return the problem (domain) to which this state belongs
	 */
	public IProblem getProblem();

	/**
	 * Returns the list of improvement operators applicable in this state. Other
	 * than getOperators() this method returns operators applicable in local
	 * search to transform a solution state into another solution state. The
	 * default implementation returns the list of all operators.
	 * @return the list of improvement operators applicable in this domain
	 */
	public List<IOperator> getImprovementOperators();

	/**
	 * Returns a random operator from the list of all applicable operators in
	 * this state. The default implementation returns a random operator from the
	 * list of all operators of the problem.
	 * @return a random operator from the list of all operators of the problem
	 */
	public IOperator getRandomOperator();

	/**
	 * Returns a random operator within a neighborhood radius passed
	 * @param neighborhoodRadius the radius of the neighborhood within which is
	 *        searched
	 * @return a random operator within a neighborhood radius passed
	 */
	public IOperator getRandomOperator(int neighborhoodRadius);

	/**
	 * Displays this state. The default implementation prints toString() to the
	 * standard ouput
	 */
	public void display();

	/**
	 * Returns an estimate for the remaining costs to get from this state to a
	 * goal state. The default implementation always returns 0.0. Overwrite this
	 * method to add heuristics information to informed search strategies.
	 * @return an estimate for the remaining costs to get from this state to a
	 *         goal state.
	 */
	public float getHeuristicsValue();

	/**
	 * Returns the utility of this state as calculated by the calculateUtility
	 * method
	 * @return the utility of this state
	 */
	public float getUtility();

	/**
	 * Sets the utility of this state. Usually used by learning agents to update
	 * utility values.
	 * @param kdo.util the new utility of this state
	 */
	public void setUtility(float util);

	/**
	 * Returns the kdo.reinforcement the agents gets when exploring this state. The
	 * default implementation allways returns 0.
	 * 
	 * @return the kdo.reinforcement the agents gets when exploring this state
	 */
	public float getReinforcement();

	/**
	 * @return the number of times this agent explored this state
	 */
	public int getExplorationCount();

	/**
	 * Called if this state is explored by the agent. The default implementation
	 * does nothing.
	 */
	public void onExploration();

	/**
	 * Checks the consistency of this state to be defined by each problem state
	 * separately
	 */
	public void checkConsistency();

}