/*
 * Copyright (c) 2008 Klaus Dorer
 * Hochschule Offenburg
 */
package kdo.domain.model;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Arrays;
import java.util.Comparator;

import kdo.domain.IIndividuum;

/**
 * Base class for all Individuums of a genetic population
 * @author dorer
 */
public abstract class Individuum extends ProblemState implements IIndividuum
{
	/** the cumulative normalized fitness share of this individuum */
	private float selectionCriteria;

	/** the number of generations this individuum is alife */
	private int age;

	/** the gene information of this individuum */
	protected float[] chromosom;

	/**
	 * Default constructor
	 */
	public Individuum(float[] chromosom)
	{
		this.chromosom = chromosom;
		age = 0;
	}

	/*
	 * (non-Javadoc)
	 * @see domain.IIndividuum#getAge()
	 */
	public int getAge()
	{
		return age;
	}

	/*
	 * (non-Javadoc)
	 * @see domain.IIndividuum#getChromosom()
	 */
	public float[] getChromosom()
	{
		return chromosom;
	}

	/*
	 * (non-Javadoc)
	 * @see domain.IIndividuum#getFitness()
	 */
	public float calculateFitness()
	{
		return calculateUtility();
	}

	/*
	 * (non-Javadoc)
	 * @see domain.IIndividuum#getFitness()
	 */
	public float getFitness()
	{
		// we have to make sure that fitness is a positive value
		return getUtility();
	}

	/*
	 * (non-Javadoc)
	 * @see domain.IIndividuum#getRelativeDiversity(domain.IIndividuum)
	 */
	public float getRelativeDiversity(IIndividuum other)
	{
		float[] chromosom1 = getChromosom();
		float[] chromosom2 = other.getChromosom();
		int sum = 0;
		for (int i = 0; i < chromosom1.length; i++) {
			if (chromosom1[i] != chromosom2[i]) {
				sum++;
			}
		}
		return sum / (float) chromosom1.length;
	}

	/*
	 * (non-Javadoc)
	 * @see domain.IIndividuum#getSelectionCriteria()
	 */
	public float getSelectionCriteria()
	{
		return selectionCriteria;
	}

	/*
	 * (non-Javadoc)
	 * @see domain.IIndividuum#incrementAge()
	 */
	public void incrementAge()
	{
		age++;
	}

	/*
	 * (non-Javadoc)
	 * @see domain.IIndividuum#setSelectionCriteria(float)
	 */
	public void setSelectionCriteria(float fitnessShare)
	{
		this.selectionCriteria = fitnessShare;
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public int compareTo(IIndividuum o)
	{
		return Float.compare(getFitness(), o.getFitness());
	}

	/**
	 * Writes this individuum to the passed file
	 * @param filename path and name of the file to write
	 * @throws IOException if writing is not possible
	 */
	public void writeToFile(String filename) throws IOException
	{
		BufferedWriter out = new BufferedWriter(new FileWriter(filename));
		out.write(Arrays.toString(chromosom));
		out.close();
	}

	/**
	 * Compares the individuums by age (ascending)
	 * @author dorer
	 */
	public static class IndividuumAgeComparator implements
		Comparator<IIndividuum>
	{
		/**
		 * @see Comparator#compare(Object, Object)
		 */
		public int compare(IIndividuum o1, IIndividuum o2)
		{
			if (o1.getAge() < o2.getAge()) {
				return -1;
			} else if (o1.getAge() > o2.getAge()) {
				return 1;
			}
			return 0;
		}
	}

	/**
	 * Compares the individuums by fitness (ascending)
	 * @author dorer
	 */
	public static class IndividuumFitnessComparator implements
		Comparator<IIndividuum>
	{
		/**
		 * @see Comparator#compare(Object, Object)
		 */
		public int compare(IIndividuum o1, IIndividuum o2)
		{
			if (o1.getFitness() < o2.getFitness()) {
				return -1;
			} else if (o1.getFitness() > o2.getFitness()) {
				return 1;
			}
			return 0;
		}
	}
}
