/*
 * Lecture symbolic AI, FH-Furtwangen
 * Original Author: Klaus Dorer
 */

package kdo.domain.model;

import java.util.List;

import kdo.domain.IOperator;
import kdo.domain.IProblem;
import kdo.domain.IProblemState;

/**
 * Represents a single state of a problem
 */
public abstract class ProblemState implements IProblemState
{
	/** The value that indicates the utility is not yet calculated */
	private final static float NOT_CALCULATED = -1.0102f;

	/**
	 * the problem domain this state belongs to. this static implementation
	 * limits the application to only treat one domain in one virtual machine.
	 * But this is what will usually be done and saves huge amount of memory.
	 */
	protected static Problem problem;

	public static long generatedStates = 0;

	/** the utility of this state */
	protected float utility;

	/** the number of times the agent visited this state */
	protected int explorationCount;

	/**
	 * Default constructor
	 */
	public ProblemState()
	{
		utility = NOT_CALCULATED;
		explorationCount = 0;
		generatedStates++;
	}

	/*
	 * (non-Javadoc)
	 * @see domain.IProblemState#getProblem()
	 */
	public IProblem getProblem()
	{
		return problem;
	}

	/**
	 * Returns false by default. Overwrite for domains with specific goal states
	 * @see kdo.domain.IProblemState#checkGoalState()
	 */
	public boolean checkGoalState()
	{
		return false;
	}

	/**
	 * @see kdo.domain.IProblemState#getOperators()
	 */
	public List<IOperator> getOperators()
	{
		return problem.getOperators();
	}

	/**
	 * @see kdo.domain.IProblemState#getImprovementOperators()
	 */
	public List<IOperator> getImprovementOperators()
	{
		return problem.getOperators();
	}

	/**
	 * @see kdo.domain.IProblemState#getRandomOperator()
	 */
	public IOperator getRandomOperator()
	{
		return getRandomOperator(1);
	}

	/*
	 * (non-Javadoc)
	 * @see domain.IProblemState#getRandomOperator(int)
	 */
	public IOperator getRandomOperator(int neighborhoodRadius)
	{
		return problem.getRandomOperator(neighborhoodRadius);
	}

	/**
	 * @see kdo.domain.IProblemState#display()
	 */
	public void display()
	{
		System.out.println(toString());
	}

	/**
	 * @see kdo.domain.IProblemState#getHeuristicsValue()
	 */
	public float getHeuristicsValue()
	{
		return 0.0f;
	}

	/**
	 * @see kdo.domain.IProblemState#calculateUtility()
	 */
	protected float calculateUtility()
	{
		return (utility = 0.0f);
	}

	/**
	 * @see kdo.domain.IProblemState#getUtility()
	 */
	public float getUtility()
	{
		// make sure to calculate the utility of a state only once
		if (utility == NOT_CALCULATED) {
			return calculateUtility();
		}
		return utility;
	}

	/**
	 * @see kdo.domain.IProblemState#setUtility(float)
	 */
	public void setUtility(float util)
	{
		utility = util;
	}

	/**
	 * @see kdo.domain.IProblemState#getReinforcement()
	 */
	public float getReinforcement()
	{
		return 0.0f;
	}

	/**
	 * @see kdo.domain.IProblemState#getExplorationCount()
	 */
	public int getExplorationCount()
	{
		return explorationCount;
	}

	/**
	 * @see kdo.domain.IProblemState#onExploration()
	 */
	public void onExploration()
	{
		explorationCount++;
	}

	/**
	 * @param problem the problem this state belongs to. The problem should not
	 *        change during runtime
	 */
	public static void setProblem(Problem problem)
	{
		assert ProblemState.problem == null || ProblemState.problem == problem : "No problem switch allowed";
		ProblemState.problem = problem;
	}

	/*
	 * (non-Javadoc)
	 * @see domain.IProblemState#checkConsistency()
	 */
	public void checkConsistency()
	{
		// the default implementation does nothing
	}
}
