/*
 * Copyright (c) 2007 Klaus Dorer
 * Hochschule Offenburg
 */
package kdo.search;

import java.util.Arrays;

import kdo.domain.IGeneticProblem;
import kdo.domain.IIndividuum;
import kdo.domain.logistics.networkOpt.HubConsolidation;
import kdo.search.strategy.local.genetic.IMutation;
import kdo.search.strategy.local.genetic.IPopulation;
import kdo.search.strategy.local.genetic.IReproduction;
import kdo.search.strategy.local.genetic.ISelection;
import kdo.search.strategy.local.genetic.mutation.RandomMutation;
import kdo.search.strategy.local.genetic.representation.Population;
import kdo.search.strategy.local.genetic.reproduction.IntMultiCrossover;
import kdo.search.strategy.local.genetic.selection.MonteCarloSelection;
import kdo.search.strategy.local.genetic.selection.SelectionStrategy;
import kdo.util.IRandomSource;
import kdo.util.RandomSource;

/**
 * @author klaus
 * 
 */
public class GeneticMain
{

	/**
	 * @param args
	 */
	public static void main(String[] args)
	{
		final int MAX_RUNTIME = 60000;
		final int GENERATIONS_COUNT = 10000;
		final int POPULATION_SIZE = 300;
		final int GENDER_SIZE = 4;
		final int OLD_TO_NEW_GENERATION_COUNT = 10;
		final float INDIVIDUUM_MUTATION_PROBABILITY = 0.05f; // 0.0 to switch off
		final float GENE_MUTATION_PROBABILITY = 0.05f;
		final int PARENTS_PER_INDIVIDUUM = 2;

		String orderFile = "data/logistics/transportAmounts.txt";
		String operatorFile = "data/logistics/possiblePaths.txt";

		IRandomSource randomSource = new RandomSource(System.currentTimeMillis());
		// IRandomSource randomSource = new RandomSource(10);

		// create the problem instance
		IGeneticProblem domain = HubConsolidation.getInstance(orderFile,
			operatorFile, randomSource);
		HubConsolidation.setLocalSearchMode();

		// create the reproduction strategy
		// IReproduction reproductionStrategy = new
		// IntSingleCrossover(randomSource);
		IReproduction reproductionStrategy = new IntMultiCrossover(randomSource);

		// create the selection strategy
		// ISelection selectionStrategy = new EliteSelection(randomSource);
		ISelection selectionStrategy = new MonteCarloSelection(false,
			randomSource);
		// ISelection selectionStrategy = new MinDiversityMonteCarloSelection(
		// randomSource, 0.1f, 10, false);

		// create the mutation strategy
		IMutation mutationStrategy = new RandomMutation(
			INDIVIDUUM_MUTATION_PROBABILITY, GENE_MUTATION_PROBABILITY,
			randomSource);

		IPopulation population = new Population(domain, GENDER_SIZE,
			POPULATION_SIZE, selectionStrategy, reproductionStrategy,
			mutationStrategy, OLD_TO_NEW_GENERATION_COUNT, PARENTS_PER_INDIVIDUUM,
			null);

		// search for solution
		long startTime = System.currentTimeMillis();

		IIndividuum best = population.breed(GENERATIONS_COUNT, MAX_RUNTIME);

		long stopTime = System.currentTimeMillis();
		long runtime = stopTime - startTime;
		System.out.println("best result: " + best);
		System.out.println("best result: " + best.getFitness());
		System.out.println("Time for search: " + runtime);

		System.out
			.println("Selected: "
				+ Arrays
					.toString(((SelectionStrategy) selectionStrategy).sumSelected));
		System.out.println("Avg fitness: "
			+ Arrays
				.toString(((SelectionStrategy) selectionStrategy).averageFitness));

		StringBuffer result = new StringBuffer(500);
		result
			.append("km, time, diversity, maxruntime, generations, population, genders, oldToNew, imp, gmp, repro, selection, mutation\n");
		result.append(best.getFitness()).append(", ");
		result.append(runtime).append(", ");
		result.append(population.getAverageDiversity()).append(", ");
		result.append(MAX_RUNTIME).append(", ");
		result.append(GENERATIONS_COUNT).append(", ");
		result.append(POPULATION_SIZE).append(", ");
		result.append(GENDER_SIZE).append(", ");
		result.append(OLD_TO_NEW_GENERATION_COUNT).append(", ");
		result.append(INDIVIDUUM_MUTATION_PROBABILITY).append(", ");
		result.append(GENE_MUTATION_PROBABILITY).append(", ");
		result.append(reproductionStrategy.getName()).append(", ");
		result.append(selectionStrategy.getName()).append(", ");
		result.append(mutationStrategy.getName()).append(", ");

		System.out.println(result.toString());

		best.checkConsistency();
	}
}
