/*
 * Lecture symbolic AI, FH-Furtwangen
 * Original Author: Klaus Dorer
 */
package kdo.search;

// import java.io.FileOutputStream;
// import java.io.IOException;
// import java.io.PrintStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Arrays;
import java.util.StringTokenizer;

import kdo.domain.IGeneticProblem;
import kdo.domain.IProblemState;
import kdo.domain.logistics.networkOpt.HubConsolidation;
import kdo.domain.logistics.networkOpt.LogisticsState;
import kdo.domain.model.ProblemState;
import kdo.search.impl.SearchParameters;
import kdo.search.strategy.ILocalSearchStrategy;
import kdo.search.strategy.local.BasicVariableNeighborhoodSearch;
import kdo.search.strategy.local.HillClimbing;
import kdo.search.strategy.local.HillClimbingRandom;
import kdo.search.strategy.local.HillClimbingRandomRestart;
import kdo.search.strategy.local.RandomWalk;
import kdo.search.strategy.local.TabuSearch;
import kdo.search.strategy.local.annealing.AnnealingAndClimbing;
import kdo.search.strategy.local.annealing.LinearTemperatureReduction;
import kdo.search.strategy.local.annealing.SigmoidalTemperatureReduction;
import kdo.search.strategy.local.annealing.SimulatedAnnealing;
import kdo.search.strategy.local.genetic.impl.GeneticSearch;
import kdo.search.strategy.local.genetic.impl.GeneticSearchParameter;
import kdo.search.strategy.local.genetic.mutation.RandomMutation;
import kdo.search.strategy.local.genetic.reproduction.IntMultiCrossover;
import kdo.search.strategy.local.genetic.selection.MonteCarloSelection;
import kdo.util.IRandomSource;
import kdo.util.RandomSource;

/**
 * This is the main class that allows to select the domain which should be used
 * @author Klaus Dorer
 */
public class LogisticsMain
{
	/** the parameters needed for searching */
	SearchParameters params;

	/**
	 * The main method to start the search or two player game selected
	 * @param args the list of command line options
	 */
	public static void main(String[] args)
	{
		if (args.length > 0 && args[0].equals("-help")) {
			printUsage();
		}
		LogisticsMain instance = new LogisticsMain();
		instance.run(args);
	}

	/**
	 * Prints a short help for command line arguments
	 */
	private static void printUsage()
	{
		System.out.println("strategy [repeats [orderFile [operatorFile]]]"
			+ "\n 1 - HillClimbing"
			+ "\n 2 - HillClimbing with Random Start state"
			+ "\n 3 - HillClimbing with Random Operator selection"
			+ "\n 4 - Random Walk                                        "
			+ "\n 10 - Tabu Search                                       "
			+ "\n 20 - Simulated Annealing"
			+ "\n 21 - Simulated Annealing with HillClimbing"
			+ "\n 30 - Basic Variable Neighborhood Search"
			+ "\n 40 - Genetic Search");
	}

	public void run(String[] args)
	{
		String strategy = "10";
		if (args != null) {
			if (args.length > 0) {
				strategy = args[0];
			}
		}

		IRandomSource randomSource = new RandomSource(System.currentTimeMillis());
		if (strategy.equals("40")) {
			params = new GeneticSearchParameter(randomSource);

		} else {
			params = new SearchParameters(randomSource, strategy);
		}

		String orderFile = "data/logistics/transportAmounts.txt";
		String operatorFile = "data/logistics/possiblePaths.txt";

		if (args != null) {
			if (args.length > 1) {
				params.setRepeats((new Integer(args[1])).intValue());
			}
			if (args.length > 2) {
				orderFile = args[2];
			}
			if (args.length > 3) {
				operatorFile = args[3];
			}
		}

		IGeneticProblem problem = HubConsolidation.getInstance(orderFile,
			operatorFile, params.getRandomSource());
		params.setDomain(problem);
		HubConsolidation.setLocalSearchMode();

		int start = 1;
		int stop = 2;
		double[][] allResults = new double[stop - start][];

		// run the series of tests
		for (int var = start; var < stop; var++) {
			float param = var * 0.1f;

			allResults[var - start] = runTests(params);
			allResults[var - start][4] = param;
		}

		// print once again all results
		System.out.println("gender size: avg, best, stddev, avgruntime, param");
		for (double[] row : allResults) {
			System.out.println(Arrays.toString(row));
		}
	}

	private double[] runTests(SearchParameters params)
	{
		TestResult result;
		double averageUtility = 0.0;
		long sumRuntime = 0;
		double bestUtility = Double.POSITIVE_INFINITY;
		long statesGenerated = ProblemState.generatedStates;
		double variance = 0.0;
		double newAvg;
		double[] resultValues = new double[5];
		for (int i = 0; i < params.getRepeats(); i++) {
			System.out.println("Run " + (i + 1) + " of " + params.getRepeats());
			result = runTest(params);
			double utility = -result.resultState.getUtility();
			newAvg = (averageUtility * i + utility) / (i + 1);
			if (i > 0) {
				variance = (i + 1) * (newAvg - averageUtility)
					* (newAvg - averageUtility) + (i - 1) * variance / i;
			}

			averageUtility = newAvg;
			sumRuntime += result.runtime;
			if (utility < bestUtility) {
				bestUtility = utility;
			}
		}

		long avgStates = (ProblemState.generatedStates - statesGenerated)
			/ params.getRepeats();
		double stddev = Math.sqrt(variance);
		long avgRuntime = sumRuntime / params.getRepeats() / 1000;
		System.out.println("Runs: " + params.getRepeats() + ", Average utility: "
			+ averageUtility + ", Best utility: " + bestUtility + " stddev: "
			+ stddev + " Average runtime: " + avgRuntime
			+ " Average Explored states: " + avgStates);
		System.out.println("------------------------------");
		resultValues[0] = averageUtility;
		resultValues[1] = bestUtility;
		resultValues[2] = stddev;
		resultValues[3] = avgRuntime;
		return resultValues;
	}

	private TestResult runTest(SearchParameters params)
	{
		// first do solution generation
		IProblemState resultState = createStartState(params);

		// now change to local search
		ILocalSearchStrategy localSearchStrategy = getStrategy(params);

		// do the local search
		System.out.println("Starting " + localSearchStrategy.getName());
		System.out.println("Utility at start: " + resultState.getUtility());
		long startTime = System.currentTimeMillis();

		resultState = localSearchStrategy.search(resultState);

		long runtime = System.currentTimeMillis() - startTime;

		// print the solution time and space
		// System.out.println("result state: " + resultState);
		// System.out.println("Time for search: " + (stopTime - startTime));
		// System.out.println("Nodes created: " + Node.getNumberOfNodes());

		StringBuffer result = new StringBuffer(500);
		result.append("km, time, strategy\n");
		result.append(-resultState.getUtility()).append(", ");
		result.append(runtime).append(", ");
		result.append(localSearchStrategy.getName()).append(", ");
		System.out.println(result);

		resultState.checkConsistency();

		storeResult(resultState);

		return new TestResult(resultState, runtime);
	}

	/**
	 * @param params optimization parameters
	 * @return the state to start the search
	 */
	private IProblemState createStartState(SearchParameters params)
	{
		IProblemState resultState;

		if (true) {
			// we want to have a trivial solution doing all orders direct path
			resultState = new LogisticsState(new float[LogisticsState.orders
				.size()]);
		} else {
			resultState = createFilteredStartState();
		}

		params.getDomain().setCurrentState(resultState);
		// System.out.println("initial solution: " + resultState.getUtility());
		return resultState;
	}

	/**
	 * Reads a start state from a file containing the relay locations
	 * @return the start state to use for search
	 */
	private IProblemState createFilteredStartState()
	{
		LogisticsState result;
		String filename = "data/logistics/relayFilter.txt";
		int size = LogisticsState.orders.size();
		float[] relay = new float[size];
		try {
			BufferedReader in = new BufferedReader(new FileReader(filename));
			String locations = in.readLine();
			StringTokenizer tokenizer = new StringTokenizer(locations, ",[]");
			// int quality = new Integer(tokenizer.nextToken());
			// tokenizer.nextToken();

			for (int i = 0; i < size; i++) {
				relay[i] = new Integer(tokenizer.nextToken().trim());
			}

			in.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		result = new LogisticsState(relay);
		return result;
	}

	/**
	 * Writes the order relay information to a file
	 * @param resultState the state for which to write
	 */
	private void storeResult(IProblemState resultState)
	{
		LogisticsState result = (LogisticsState) resultState;
		String filename = "generesult.txt";
		try {
			BufferedWriter out = new BufferedWriter(new FileWriter(filename, true));
			out.write(result.getUtility() + ", "
				+ result.getRelayInformationString() + "\n");
			out.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Returns the search strategy as defined by the passed command line strategy
	 * string
	 * 
	 * @param params the parameters used to set up search strategy
	 * @return the local search strategy defined by the parameter passed
	 */
	private ILocalSearchStrategy getStrategy(SearchParameters params)
	{
		ILocalSearchStrategy localSearchStrategy;
		if (params.getStrategy().equals("1")) {
			// do hill climbing
			localSearchStrategy = new HillClimbing(-1);

		} else if (params.getStrategy().equals("2")) {
			// do random operator selection
			localSearchStrategy = new HillClimbingRandomRestart(10, -1);

		} else if (params.getStrategy().equals("3")) {
			// do random operator selection
			localSearchStrategy = new HillClimbingRandom(1000000, params
				.getMaxRuntime());

		} else if (params.getStrategy().equals("4")) {
			// do random walk
			localSearchStrategy = new RandomWalk(30, 100000, params
				.getMaxRuntime());

		} else if (params.getStrategy().equals("20")) {
			// do simulated annealing
			localSearchStrategy = new SimulatedAnnealing(
				new SigmoidalTemperatureReduction(100.0, 0.0), 1000000, -1);
			// new LinearTemperatureReduction(30.0, 0.0), 1000000, -1);

		} else if (params.getStrategy().equals("21")) {
			// do simulated annealing and hill climbing
			localSearchStrategy = new AnnealingAndClimbing(
				new LinearTemperatureReduction(40.0, 0.0), 500000, -1);

		} else if (params.getStrategy().equals("30")) {
			// do basic variable neighborhood search
			localSearchStrategy = new HillClimbing(-1);
			localSearchStrategy = new BasicVariableNeighborhoodSearch(
				localSearchStrategy, 175, params.getMaxRuntime());

		} else if (params.getStrategy().equals("40")) {
			// do genetic search
			GeneticSearchParameter parameters = (GeneticSearchParameter) params;
			parameters.setReproductionStrategy(new IntMultiCrossover(parameters
				.getRandomSource()));
			parameters.setSelectionStrategy(new MonteCarloSelection(false,
				parameters.getRandomSource()));
			parameters.setMutationStrategy(new RandomMutation(parameters
				.getIndividuumMutationProbability(), parameters
				.getGeneMutationProbability(), parameters.getRandomSource()));

			localSearchStrategy = new GeneticSearch(parameters);

		} else {
			// do tabu search
			localSearchStrategy = new TabuSearch(100, 100000, params
				.getMaxRuntime());
		}
		return localSearchStrategy;
	}

	class TestResult
	{
		IProblemState resultState;

		long runtime;

		/**
		 * @param resultState
		 * @param runtime
		 */
		public TestResult(IProblemState resultState, long runtime)
		{
			super();
			this.resultState = resultState;
			this.runtime = runtime;
		}
	}
}
