/*
 * Lecture symbolic AI, FH-Furtwangen
 * Original Author: Klaus Dorer
 */

package kdo.search.strategy.local;

import java.util.List;

import kdo.domain.IOperator;
import kdo.domain.IProblemState;


/**
 * Implementation of hill climbing local search
 */
public class HillClimbing extends LocalSearchBase
{
	/**
	 * Constructor
	 * @param maxRuntime the maximal runtime the search strategy may use for a
	 *        single call to search, -1 if no time restriction
	 */
	public HillClimbing(long maxRuntime)
	{
		super("HillClimbing", null, maxRuntime);
	}

	/**
	 * Local search algorithm for hill climbing that does a steepest gradient
	 * ascend.
	 * @param initialState the state to start the search from
	 * @return the last state in which the search ended
	 */
	public IProblemState search(IProblemState initialState)
	{
		// perform the search
		IOperator operator;
		IProblemState successorState;
		IProblemState bestSuccessorState = initialState;
		IProblemState currentState = initialState;
		int iterations = 0;

		onStartSearch();
		while (true) {
			// get a list of all directions (operators) where to go
			List<IOperator> operators = getNeighborhood(currentState);
			int size = operators.size();

			for (int i = 0; i < size; i++) {
				operator = operators.get(i);
				// calculate the utility if we go this way
				successorState = operator.getSuccessorState(currentState);
				if (successorState.getUtility() > bestSuccessorState.getUtility()) {
					// we have a steeper gradient
					bestSuccessorState = successorState;
				}
			}

			printState(currentState, iterations);

			if (canStopHillClimbing(bestSuccessorState, currentState, iterations)) {
				// we have reached a (local) maximum
				return currentState;
			}

			currentState = bestSuccessorState;
			iterations++;
		}
	}

	/**
	 * Checks if the hill climbing algorithm can be stopped. The default
	 * implementation checks if the best state to go is the same state where we
	 * are.
	 * 
	 * @param bestSuccessorState the best state of the neighborhood to which to
	 *        go
	 * @param currentState the current state where we are
	 * @param iterations the number of iterations the algorithm went through
	 *        since call of search()
	 * @return true if the algorithm can be stopped
	 */
	protected boolean canStopHillClimbing(IProblemState bestSuccessorState,
			IProblemState currentState, int iterations)
	{
		if (bestSuccessorState == currentState) {
			return true;
		}

		return super.canStop();
	}
}
