/*
 * Lecture symbolic AI, FH-Furtwangen
 * Original Author: Klaus Dorer
 */
package kdo.search.strategy.local;

import java.util.Collections;
import java.util.List;

import kdo.domain.IOperator;
import kdo.domain.IProblemState;


/**
 * Represents a hill climbing local search where not the full neighborhood is
 * explored, but only one random direction. If that improves the current state,
 * we take it, if not we do not. The search stops if maxCyclesWithoutImprovement
 * cycles did not produce an improvement.
 * 
 * @author Last modified by $Author: KDorer $
 * @version $Revision: 1.3 $
 */
public class HillClimbingRandom extends HillClimbing
{
	/** the number of cycles that did not produce an improvement */
	private int cyclesWithoutImprovement;

	/** the maximum number of cycles that did not produce an improvement */
	private final int maxCyclesWithoutImprovement;

	/**
	 * Creates an instance that does only a limited amount of cycles without
	 * improvement before it stops.
	 * @param maxCyclesWithoutImprovement the
	 * @param maxRuntime the maximal runtime the search strategy may use for a
	 *        single call to search, -1 if no time restriction
	 */
	public HillClimbingRandom(int maxCyclesWithoutImprovement, long maxRuntime)
	{
		super(maxRuntime);

		cyclesWithoutImprovement = 0;
		this.maxCyclesWithoutImprovement = maxCyclesWithoutImprovement;
	}

	/*
	 * (non-Javadoc)
	 * @see search.strategy.IStrategy#getName()
	 */
	@Override
	public String getName()
	{
		return "HillClimbing Random Operator. maxCyclesWithoutImprovement: "
				+ maxCyclesWithoutImprovement;
	}

	/**
	 * Returns a single random operator of the neighborhood. If too many cycles
	 * did not produce an improvement, all operators for normal hill climbing are
	 * returned.
	 * 
	 * @param currentState the state of the domain for which to get the operators
	 * @return a single random operator of the neighborhood
	 * 
	 * @see kdo.search.strategy.local.HillClimbing#getNeighborhood(kdo.domain.Problem,
	 *      kdo.domain.ProblemState)
	 */
	@Override
	protected List<IOperator> getNeighborhood(IProblemState currentState)
	{
		if (cyclesWithoutImprovement > maxCyclesWithoutImprovement) {
			// after too many cycles without improvement we switch to normal hill
			// climbing
			return super.getNeighborhood(currentState);
		} else {
			return Collections.singletonList(currentState.getRandomOperator());
		}
	}

	/**
	 * Checks if an improvement was made or if the number of non-improving cycles
	 * is low enough
	 * 
	 * @param bestSuccessorState the best state of the neighborhood to which to
	 *        go
	 * @param currentState the current state where we are
	 * @param iterations the number of iterations the algorithm went through
	 *        since call of search()
	 * @return false if an improvement was made or if the number of non-improving
	 *         cycles is low enough
	 * 
	 * @see kdo.search.strategy.local.HillClimbing#reachedLocalMaximum(kdo.domain.ProblemState,
	 *      kdo.domain.ProblemState, int)
	 */
	@Override
	protected boolean canStopHillClimbing(IProblemState bestSuccessorState,
			IProblemState currentState, int iteration)
	{
		if (cyclesWithoutImprovement > maxCyclesWithoutImprovement) {
			// too many cycles without improvement, so normal hillclimbing
			return super.canStopHillClimbing(bestSuccessorState, currentState,
					iteration);
		}

		if (bestSuccessorState.getUtility() > currentState.getUtility()) {
			// we have an improvement
			cyclesWithoutImprovement = 0;
			return false;
		}
		cyclesWithoutImprovement++;
		return canStop();
	}
}
