/*
 * Lecture symbolic AI, FH-Furtwangen
 * Original Author: Klaus Dorer
 */

package kdo.search.strategy.local;

import java.util.List;

import kdo.domain.IOperator;
import kdo.domain.IProblemState;
import kdo.search.impl.StrategyBase;
import kdo.search.strategy.ILocalSearchStrategy;
import kdo.util.IRandomSource;


/**
 * Implementation of hill climbing local search
 */
public abstract class LocalSearchBase extends StrategyBase implements
		ILocalSearchStrategy
{
	/** the maximal runtime to run in ms */
	long maximalRuntime;

	/** the time of the last call to search */
	long searchStarttime;

	/** time of last call to printState */
	long lastprintStateTime;

	/**
	 * Constructor
	 * @param name the name of this strategy
	 * @param rand source for pseudo random numbers
	 * @param maxRuntime the maximal runtime the search strategy may use for a
	 *        single call to search, -1 if no time restriction
	 */
	public LocalSearchBase(String name, IRandomSource rand, long maxRuntime)
	{
		super(name, rand);
		maximalRuntime = maxRuntime;
		if (maximalRuntime <= 0) {
			maximalRuntime = Long.MAX_VALUE;
		}
		lastprintStateTime = 0;
	}

	protected void onStartSearch()
	{
		searchStarttime = System.currentTimeMillis();
	}

	/**
	 * Returns a list of operators that lead to the neighborhood states of the
	 * state passed. The default implementation returns all possible operators
	 * 
	 * @param currentState the state of the domain for which to get the operators
	 * @return a list of operators that lead to the neighborhood states of the
	 *         state passed
	 */
	protected List<IOperator> getNeighborhood(IProblemState currentState)
	{
		// get all possible state transition operators
		return currentState.getImprovementOperators();
	}

	/**
	 * @return the runtime since last call to onStartSearch
	 */
	protected long getRuntime()
	{
		return System.currentTimeMillis() - searchStarttime;
	}

	/**
	 * Checks if the search algorithm can be stopped. The default implementation
	 * checks if the specified max runtime is used up.
	 * 
	 * @return true if the algorithm can be stopped
	 */
	protected boolean canStop()
	{
		long runtime = getRuntime();
		if (runtime > maximalRuntime) {
			return true;
		}

		return false;
	}

	/**
	 * @return the remaining runtime in seconds
	 */
	protected long getRemainingRuntime()
	{
		return (maximalRuntime - getRuntime()) / 1000;
	}

	/**
	 * @param state the current state of searching
	 * @param iterations the number of iterations done so far
	 * @return true if we printed, false if not
	 */
	protected boolean printState(IProblemState state, long iterations)
	{
		long nowTime = System.currentTimeMillis();
		long timeDelta = nowTime - lastprintStateTime;
		if (timeDelta > 1000) {
			String output = iterations + " utility: " + state.getUtility();
			if (maximalRuntime < Long.MAX_VALUE) {
				output += " time to go: " + getRemainingRuntime();
			} else {
				output += " time running: " + (getRuntime() / 1000);
			}
			System.out.println(output);
			lastprintStateTime = nowTime;
			return true;
		}
		return false;
	}
}
