/*
 * Lecture symbolic AI, FH-Furtwangen
 * Original Author: Klaus Dorer
 */

package kdo.search.strategy.local;

import java.util.LinkedList;
import java.util.List;

import kdo.domain.IOperator;
import kdo.domain.IProblemState;

/**
 * Implementation of local search with tabu list for cycle prevention. In order
 * to work you have to make sure that you have implemented a meaningfull equals
 * method for your domain's state class!
 */
public class TabuSearch extends LocalSearchBase
{
	/** the size of the short term memory tabu list */
	private final int shortTermTabuSize;

	/** the short term tabu list to store states that are tabu */
	private final List<IProblemState> shortTermTabuList;

	/** the maximal number of walk steps to do */
	private final int maxIterations;

	/**
	 * Constructor setting the size of the tabu list
	 * @param shortTermTabuSize the size of the short term tabu list
	 * @param maxIterations the maximal number of iterations done before stopping
	 * @param maxRuntime the maximal runtime the search strategy may use for a
	 *        single call to search, -1 if no time restriction
	 */
	public TabuSearch(int shortTermTabuSize, int maxIterations, long maxRuntime)
	{
		super("Tabu Search. shortTermTabuSize: " + shortTermTabuSize
			+ " maxIterations: " + maxIterations, null, maxRuntime);
		this.shortTermTabuSize = shortTermTabuSize;
		shortTermTabuList = new LinkedList<IProblemState>();
		this.maxIterations = maxIterations;
	}

	/**
	 * Local search algorithm for tabu search.
	 * @param initialState the state to start the search from
	 * @return the last state in which the search ended
	 */
	public IProblemState search(IProblemState initialState)
	{
		// perform the search
		IOperator operator;
		int iterations = 0;
		IProblemState successorState;
		IProblemState bestSuccessorState;
		IProblemState bestState = initialState;
		IProblemState currentState = initialState;

		onStartSearch();
		while (iterations++ < maxIterations && !canStop()) {
			// get all possible state transition operators
			List<IOperator> operators = currentState.getImprovementOperators();
			bestSuccessorState = null;
			int size = operators.size();

			for (int i = 0; i < size; i++) {
				operator = operators.get(i);
				// calculate the utility if we go this way
				successorState = operator.getSuccessorState(currentState);

				if (bestSuccessorState == null
					|| successorState.getUtility() > bestSuccessorState.getUtility()) {
					if (!shortTermTabuList.contains(successorState)) {
						// we have the steepest gradient so far that is not tabu
						bestSuccessorState = successorState;
					}
				}
			}

			// in case that no successor was found (all states run through were
			// tabu)
			// we go on with the best state so far
			if (bestSuccessorState == null) {
				bestSuccessorState = bestState;
			}

			if (bestSuccessorState.getUtility() > bestState.getUtility()) {
				// we have found a new optimum
				bestState = bestSuccessorState;
				printState(bestState, iterations);
			}

			// remember current state in the short term tabu list
			currentState = bestSuccessorState;
			addToShortTermTabuList(currentState);
		}

		return bestState;
	}

	/**
	 * Adds a problem state to the list of short termn tabu states, i.e. states
	 * that may not be visited in the next few steps
	 * @param state the state to add
	 */
	private void addToShortTermTabuList(IProblemState state)
	{
		shortTermTabuList.add(state);
		if (shortTermTabuList.size() > shortTermTabuSize) {
			shortTermTabuList.remove(0);
		}
	}
}
