/*
 * Copyright (c) 2007 Klaus Dorer
 * Hochschule Offenburg
 */
package kdo.search.strategy.local.genetic.representation;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;

import kdo.domain.IGeneticProblem;
import kdo.domain.IIndividuum;
import kdo.search.strategy.local.genetic.IMutation;
import kdo.search.strategy.local.genetic.IPopulation;
import kdo.search.strategy.local.genetic.IReproduction;
import kdo.search.strategy.local.genetic.ISelection;
import kdo.util.observer.IObserver;

/**
 * @author klaus
 * 
 */
public class Population implements IPopulation
{
	/** the diversity below which breeding is stopped */
	private static float MIN_DIVERSITY = 1.00001f;

	/** the individuums of this population, a list for each gender */
	protected List<List<IIndividuum>> individuums;

	/** the comparator to use to compare individuums */
	private final Comparator<IIndividuum> individuumComparator;

	/** selection strategy */
	private final ISelection selectionStrategy;

	/** selection strategy */
	private final IMutation mutationStrategy;

	/** the strategy for reproduction */
	private final IReproduction reproductionStrategy;

	/** the number of individuums to take over from old generation */
	protected int oldToNewGenerationCount;

	/** the application domain to which this population is related */
	private final IGeneticProblem domain;

	/** the number of generations this population has been breeded */
	private int generations;

	/** the number of parents a new individuum has */
	private final int parentsPerIndividuum;

	/** the average diversity over all individuums and genes */
	private float avgDiversity;

	/** observer to inform about state changes */
	private IObserver<IPopulation> observer;

	/** the time breeding is running now */
	private long runtime;

	/** flag indicating that an interrupt is requested */
	private boolean interrupted;

	/**
	 * @param domain the domain on which the population is breeded
	 * @param individuums a list of individuums that make up this population
	 * @param individuumComparator comparator that is used for comparing two
	 *        individuums
	 * @param selectionStrategy the strategy that does selection and
	 *        recombination
	 * @param reproductionStrategy the strategy that creates a new individuum
	 *        from the chromosoms of parents
	 * @param mutationStrategy the strategy that does mutation of individuums
	 * @param oldToNewGenerationCount the number of best individuums to take over
	 *        into next generation
	 */
	public Population(IGeneticProblem domain,
		List<List<IIndividuum>> individuums, ISelection selectionStrategy,
		IReproduction reproductionStrategy, IMutation mutationStrategy,
		float oldToNewGenerationRatio, int parentsPerIndividuum)
	{
		assert individuums.size() > 2 : "Too few individuums in the population! : "
			+ individuums.size();
		this.domain = domain;
		this.individuums = individuums;
		this.individuumComparator = domain.getIndividuumComparator();
		this.selectionStrategy = selectionStrategy;
		this.reproductionStrategy = reproductionStrategy;
		this.mutationStrategy = mutationStrategy;
		int individuumsPerGender = individuums.get(0).size() / individuums.size();
		this.oldToNewGenerationCount = (int) (individuumsPerGender * oldToNewGenerationRatio);
		this.parentsPerIndividuum = parentsPerIndividuum;
		generations = 0;
		avgDiversity = 2.0f;
		interrupted = false;
	}

	/**
	 * @param domain the domain on which the population is breeded
	 * @param numberOfGenders the number of genders into which the population is
	 *        split
	 * @param numberOfIndividuums the number of individuums the population should
	 *        contain
	 * @param selectionStrategy the strategy that does selection and
	 *        recombination
	 * @param reproductionStrategy the strategy that creates a new individuum
	 *        from the chromosoms of parents
	 * @param mutationStrategy the strategy that does mutation of individuums
	 * @param oldToNewGenerationRatio the relative amount of best individuums to
	 *        take over into next generation for each gender
	 */
	public Population(IGeneticProblem domain, int numberOfGenders,
		int numberOfIndividuums, ISelection selectionStrategy,
		IReproduction reproductionStrategy, IMutation mutationStrategy,
		float oldToNewGenerationRatio, int parentsPerIndividuum,
		IObserver<IPopulation> observer)
	{
		assert numberOfIndividuums > 2 : "Too few individuums in the population! : "
			+ numberOfIndividuums;
		this.domain = domain;
		int individuumsPerGender = numberOfIndividuums / numberOfGenders;
		this.individuums = generateInitialPopulation(numberOfGenders,
			individuumsPerGender);
		this.individuumComparator = domain.getIndividuumComparator();
		this.selectionStrategy = selectionStrategy;
		this.reproductionStrategy = reproductionStrategy;
		this.mutationStrategy = mutationStrategy;
		this.oldToNewGenerationCount = (int) (individuumsPerGender * oldToNewGenerationRatio);
		this.parentsPerIndividuum = parentsPerIndividuum;
		generations = 0;
		avgDiversity = 2.0f;
		this.observer = observer;
	}

	private List<List<IIndividuum>> generateInitialPopulation(int genders,
		int size)
	{
		List<List<IIndividuum>> result = new ArrayList<List<IIndividuum>>();
		for (int gender = 0; gender < genders; gender++) {
			List<IIndividuum> genderIndividuums = new ArrayList<IIndividuum>();
			for (int i = 0; i < size; i++) {
				genderIndividuums.add(createRandomIndividuum());
			}
			result.add(genderIndividuums);
		}
		return result;
	}

	/**
	 * Creates a randomly created individuum from the application domain this
	 * population is based on. The individuum is not added to the population.
	 * @return a randomly created individuum from the application domain this
	 *         population is based on
	 */
	public IIndividuum createRandomIndividuum()
	{
		return domain.getRandomIndividuum();
	}

	/**
	 * 
	 * @param domain the problem to solve using genetic algorithms
	 * @param iterations the number of generations to run
	 * @param maxRuntime the maximal time to breed in ms
	 * @return the best individuum in the final population
	 */
	public IIndividuum breed(int iterations, long maxRuntime)
	{
		long start = System.currentTimeMillis();
		IIndividuum bestIndividuum = null;
		runtime = 0;
		while (generations < iterations && runtime < maxRuntime
			&& avgDiversity > MIN_DIVERSITY && !interrupted) {
			calculateFitness();
			IIndividuum bestInPopulation = getBestIndividuum();
			// calculate fitness of each individuum
			if (bestIndividuum == null
				|| individuumComparator.compare(bestInPopulation, bestIndividuum) > 0) {
				bestIndividuum = bestInPopulation;
				// if (bestIndividuum.getFitness() < 35000) {
				printSituation(bestIndividuum, generations);
				// }
			}
			observer.update(this);

			// increase age of individuums
			age();

			// mutate new population with a small probability
			mutate();

			// select parents for reproduction and generate new population
			List<List<IIndividuum>> newIndividuums = new ArrayList<List<IIndividuum>>(
				individuums.size());
			takeOverGoodIndividuums(newIndividuums);
			createNewIndividuums(newIndividuums);
			individuums = newIndividuums;
			calculateAverageDiversity();

			generations++;
			runtime = System.currentTimeMillis() - start;
		}
		interrupted = false;
		printSituation(bestIndividuum, generations);
		observer.update(this);
		return bestIndividuum;
	}

	private void calculateFitness()
	{
		for (List<IIndividuum> gender : individuums) {
			for (IIndividuum currentIndividuum : gender) {
				currentIndividuum.calculateFitness();
			}
		}
	}

	/**
	 * Adds one to the age of each individuum
	 */
	private void age()
	{
		for (List<IIndividuum> indiList : individuums) {
			for (IIndividuum currentIndividuum : indiList) {
				currentIndividuum.incrementAge();
			}
		}
	}

	/**
	 * Runs mutations for all genders
	 */
	private void mutate()
	{
		for (List<IIndividuum> indiList : individuums) {
			mutationStrategy.mutate(indiList);
		}
	}

	/**
	 * Removes all individuums from the list of individuums of all genders except
	 * for the n best
	 */
	private void takeOverGoodIndividuums(List<List<IIndividuum>> newIndividuums)
	{
		for (List<IIndividuum> indiList : individuums) {
			int size = indiList.size();
			// take over n best
			Collections.sort(indiList, individuumComparator);
			newIndividuums.add(new ArrayList<IIndividuum>(indiList.subList(size
				- oldToNewGenerationCount, size)));
			// inform the selection strategy that selection will soon start
			selectionStrategy.onBeforeSelection(indiList);
		}
	}

	/**
	 * Creates new individuums for the passed gender
	 * @param gender the gender for which to create new individuums
	 */
	private void createNewIndividuums(List<List<IIndividuum>> newIndividuums)
	{
		int genders = individuums.size();
		for (int gender = 0; gender < genders; gender++) {
			List<IIndividuum> indiList = individuums.get(gender);
			List<IIndividuum> newList = newIndividuums.get(gender);
			int size = indiList.size();

			int newIndividuumsToGenerate = size - oldToNewGenerationCount;
			int generated = 0;
			// We have to make sure that also when having only 1 gender we select 2
			// parents
			IIndividuum[] parents = new IIndividuum[parentsPerIndividuum];
			while (generated < newIndividuumsToGenerate) {
				// select the parents
				for (int parentgender = 0; parentgender < parentsPerIndividuum; parentgender++) {
					List<IIndividuum> otherGender = individuums
						.get((gender + parentgender) % genders);
					parents[parentgender] = selectionStrategy.selectIndividuum(this,
						otherGender, generated, parents, parentgender);
				}
				IIndividuum child = reproductionStrategy.crossOver(parents);
				newList.add(child);
				generated++;
			}
		}
	}

	private void printSituation(IIndividuum bestIndividuum, int generations)
	{
		System.out.println("generations: " + generations + " utility: "
			+ bestIndividuum.getFitness() + " age: " + bestIndividuum.getAge()
			+ " population: " + this);
	}

	/*
	 * (non-Javadoc)
	 * @see search.strategy.local.genetic.IPopulation#getBestIndividuum()
	 */
	public IIndividuum getBestIndividuum()
	{
		return getBestIndividuum(getIndividuumComparator());
	}

	/*
	 * (non-Javadoc)
	 * @see search.strategy.local.genetic.IPopulation#getBestIndividuum()
	 */
	public IIndividuum getBestIndividuum(Comparator<IIndividuum> comparator)
	{
		IIndividuum best = null;
		for (List<IIndividuum> genderList : individuums) {
			IIndividuum bestOfGender = Collections.max(genderList, comparator);
			if (best == null || comparator.compare(bestOfGender, best) > 0) {
				best = bestOfGender;
			}
		}
		return best;
	}

	/*
	 * (non-Javadoc)
	 * @see search.strategy.local.genetic.IPopulation#getIndividuumComparator()
	 */
	public Comparator<IIndividuum> getIndividuumComparator()
	{
		return individuumComparator;
	}

	/*
	 * (non-Javadoc)
	 * @see search.strategy.local.genetic.IPopulation#getSize()
	 */
	public int getSize()
	{
		return individuums.size();
	}

	/*
	 * (non-Javadoc)
	 * @see search.strategy.local.genetic.IPopulation#getGenerations()
	 */
	public int getGenerations()
	{
		return generations;
	}

	/**
	 * @return the number of different genes for each position
	 */
	protected int[] getDiversity()
	{
		IIndividuum firstIndividuum = individuums.get(0).get(0);
		int genesize = firstIndividuum.getChromosom().length;
		int[] result = new int[genesize];

		// run through all gene positions
		for (int gene = 0; gene < genesize; gene++) {
			HashMap<Float, Float> currentMap = new HashMap<Float, Float>(
				individuums.size());

			// run through all individuums
			for (List<IIndividuum> currentGender : individuums) {
				for (IIndividuum currentIndividuum : currentGender) {
					Float key = currentIndividuum.getChromosom()[gene];
					// TODO we could also count the number of occurrences in the
					// value
					currentMap.put(key, key);
				}
			}
			result[gene] = currentMap.size();
		}

		return result;
	}

	/**
	 * @return the average diversity of all genes in the population
	 */
	public float getAverageDiversity()
	{
		return avgDiversity;
	}

	/**
	 * Calculates the average diversity
	 * @return the average diversity of all genes in the population
	 */
	public float calculateAverageDiversity()
	{
		int[] diversity = getDiversity();
		long sum = 0;
		for (int divers : diversity) {
			// calculate average
			sum += divers;
		}
		avgDiversity = sum / (float) diversity.length;
		return avgDiversity;
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString()
	{
		StringBuffer buffer = new StringBuffer(1000);
		// print all individuums
		// for (IIndividuum individuum : individuums) {
		// buffer.append(individuum.getFitness()).append(": ");
		// int[] chromosom = individuum.getChromosom();
		// for (int gene: chromosom) {
		// buffer.append(gene).append(", ");
		// }
		// buffer.append("\n");
		// }

		// print diversity information
		// int[] diversity = getDiversity();
		// long sum = 0;
		// buffer.append("diversity: ");
		// for (int divers : diversity) {
		// // calculate average
		// sum += divers;
		// buffer.append(divers).append(", ");
		// }
		// buffer.append("avg diversity: ").append(sum / (float)
		// diversity.length);
		buffer.append("avg diversity: ").append(getAverageDiversity());
		buffer.append(" genders: ").append(individuums.size());
		buffer.append(" individuums: ").append(
			individuums.get(0).size() * individuums.size());
		buffer.append(" parentsPerIndividuum: ").append(parentsPerIndividuum);
		buffer.append(" oldToNewGenerationCount: ").append(
			oldToNewGenerationCount);

		// System.out
		// .println("Selected: "
		// + Arrays
		// .toString(((SelectionStrategy) selectionStrategy).sumSelected));
		// System.out
		// .println("Avg fitness: "
		// + Arrays
		// .toString(((SelectionStrategy) selectionStrategy).averageFitness));

		return buffer.toString();
	}

	/**
	 * @return the time that breeding has taken until now
	 */
	public long getRuntime()
	{
		return runtime;
	}

	public void interrrupt()
	{
		interrupted = true;
	}
}
