/*
 * Lecture symbolic AI, FH-Furtwangen
 * Original Author: Klaus Dorer
 */
package kdo.search;

import java.util.List;

import kdo.domain.IProblem;
import kdo.domain.IProblemState;
import kdo.domain.approximation.FunctionProblem;
import kdo.domain.logistics.networkOpt.HubConsolidation;
import kdo.domain.logistics.networkOpt.LogisticsState;
import kdo.domain.logistics.tsp.TSPProblem;
import kdo.domain.puzzleOfN.PuzzleOfN;
import kdo.domain.tictactoe.TicTacToe;
import kdo.domain.vacuumworld.VacuumWorld;
import kdo.domain.vierGewinnt.ComputerPlayer;
import kdo.domain.vierGewinnt.Player;
import kdo.domain.vierGewinnt.VierGewinnt;
import kdo.search.impl.GeneralTwoPlayerSearch;
import kdo.search.representation.Node;
import kdo.search.strategy.ILocalSearchStrategy;
import kdo.search.strategy.IStrategy;
import kdo.search.strategy.ITreeSearchStrategy;
import kdo.search.strategy.ITwoPlayerSearchStrategy;
import kdo.search.strategy.informed.AStarSearch;
import kdo.search.strategy.informed.AStarSearchSpaceOptimized;
import kdo.search.strategy.informed.GreedySearch;
import kdo.search.strategy.local.BasicVariableNeighborhoodSearch;
import kdo.search.strategy.local.HillClimbing;
import kdo.search.strategy.local.HillClimbingRandom;
import kdo.search.strategy.local.HillClimbingRandomRestart;
import kdo.search.strategy.local.RandomWalk;
import kdo.search.strategy.local.TabuSearch;
import kdo.search.strategy.local.annealing.LinearTemperatureReduction;
import kdo.search.strategy.local.annealing.SimulatedAnnealing;
import kdo.search.strategy.local.genetic.impl.GeneticSearch;
import kdo.search.strategy.local.genetic.impl.GeneticSearchParameter;
import kdo.search.strategy.local.genetic.mutation.RandomMutation;
import kdo.search.strategy.local.genetic.reproduction.IntMultiCrossover;
import kdo.search.strategy.local.genetic.selection.MonteCarloSelection;
import kdo.search.strategy.twoPlayer.AlphaBetaSearch;
import kdo.search.strategy.twoPlayer.MiniMaxSearch;
import kdo.search.strategy.uninformed.BestFirstSearch;
import kdo.search.strategy.uninformed.BreadthFirstSearch;
import kdo.search.strategy.uninformed.DepthFirstSearch;
import kdo.search.strategy.uninformed.DepthLimitedSearch;
import kdo.search.strategy.uninformed.IterativeDeepeningSearch;
import kdo.tutorials.vacuumWorld.AStarSearchWork;
import kdo.util.IRandomSource;
import kdo.util.RandomSource;

/**
 * This is the main class that allows to select the domain which should be used
 * @author Klaus Dorer
 */
public class SearchMain
{
	/**
	 * The main method to start the search or two player game selected
	 * @param args the list of command line options
	 */
	public static void main(String[] args)
	{
		// check command line parameters
		if (args.length < 2) {
			// <todo> create gui version
			printUsage();
			System.exit(0);
		}

		String domainString = args[0];
		String strategyString = args[1];
		int paramString = 2;
		if (domainString.equals("1") || domainString.equals("21")) {
			strategyString = args[2];
			paramString = 3;
		}
		float domainParameter = new Double(args[1]).floatValue();
		float parameter1 = 0.0f;
		float parameter2 = 0.0f;
		if (args.length > paramString) {
			parameter1 = new Double(args[paramString]).floatValue();
			if (args.length > paramString + 1) {
				parameter2 = new Double(args[paramString + 1]).floatValue();
			}
		}

		// get the strategy
		IRandomSource randomSource = new RandomSource(10);
		IStrategy strategy = getStrategy(strategyString, parameter1, parameter2,
			randomSource);

		// get the domain
		IProblem domain = getDomain(domainString, domainParameter, strategy,
			randomSource);

		long startTime = System.currentTimeMillis();

		// now the call to start the game
		List<Node> result = null;
		if (strategy instanceof ITreeSearchStrategy) {
			// result = new TreeSearch().breadthFirstSearch(domain);
			result = ((ITreeSearchStrategy) strategy).search(domain
				.getInitialState());

		} else if (strategy instanceof ITwoPlayerSearchStrategy) {
			ITwoPlayerSearchStrategy strategy1 = (ITwoPlayerSearchStrategy) strategy;
			// TODO WARNING: this only works for the two stateless strategies
			// Once we have moved the strategies into the player instances of two
			// player
			// games this will be obsolete.
			ITwoPlayerSearchStrategy strategy2 = (ITwoPlayerSearchStrategy) strategy;
			GeneralTwoPlayerSearch.playGame(domain, strategy1, strategy2);

		} else if (strategy instanceof ILocalSearchStrategy) {
			// for local search we first have to find an initial solution
			System.out.println("Starting Solution Generation ...");
			IProblemState resultState = domain.getInitialState();
			// searchStrategy = new AStarSearchSpaceOptimized();
			// result = searchStrategy.search(domain.getInitialState());
			// if (result != null) {
			// Node resultNode = result.get(result.size() - 1);
			// IProblemState resultState = resultNode.getState();
			// System.out.println("result of A*: "
			// + resultNode.getState().getUtility());
			//
			// // now we can improve it
			System.out.println("Starting Local Search ...");
			resultState = ((ILocalSearchStrategy) strategy).search(resultState);
			System.out.println("Solution: " + resultState.getUtility() + "\n"
				+ resultState);
			resultState.checkConsistency();

			// !!! it is not nice that we mention the concrete problem here
			// PlanVO plan = ((LogisticsState) resultState)
			// .createTransportPlanValueObject();
			// }
		} else {
			System.out.println("Unknown strategy");
			System.exit(-1);
		}

		long stopTime = System.currentTimeMillis();

		// print the solution time and space
		if (result != null) {
			if (!result.isEmpty()) {
				Node lastNode = result.get(result.size() - 1);
				if (lastNode.getDepth() < 50) {
					lastNode.printSolutionPath();
				} else {
					System.out.println("Last node: " + lastNode);
				}
			}
		} else {
			System.out.println("No solution found.");
		}

		System.out.println("Time for search: " + (stopTime - startTime));
		System.out.println("Nodes created: " + Node.getNumberOfNodes());
		System.out.println("Nodes explored: " + (Node.getExploredNodes() + 1));
		if (strategy instanceof ITreeSearchStrategy) {
			System.out.println("Max nodes in store (search only): "
				+ ((ITreeSearchStrategy) strategy).getMaxNumberOfNodesStored());
		}
	}

	/**
	 * @param args
	 * @param domain
	 * @param domainString
	 * @param strategy
	 * @return
	 */
	private static IProblem getDomain(String domainString,
		float domainParameter, IStrategy strategy, IRandomSource randomSource)
	{
		if (domainString.equals("1")) {
			return VacuumWorld.getInstance((int) domainParameter);

		} else if (domainString.equals("2")) {
			return PuzzleOfN.getInstance(3);

		} else if (domainString.equals("20")) {
			return TicTacToe.getInstance();

		} else if (domainString.equals("21")) {
			Player player1 = new ComputerPlayer("player 1", (byte) 1,
				(ITwoPlayerSearchStrategy) strategy);
			Player player2 = new ComputerPlayer("player 2", (byte) 2,
				(ITwoPlayerSearchStrategy) strategy);
			return VierGewinnt
				.getInstance((int) domainParameter, player1, player2);

		} else if (domainString.equals("10")) {
			String orderFile = "data/logistics/transportAmounts.txt";
			String operatorFile = "data/logistics/possiblePaths.txt";
			IProblem domain = HubConsolidation.getInstance(orderFile,
				operatorFile, randomSource);
			IProblemState initial = new LogisticsState(
				new float[LogisticsState.orders.size()]);
			domain.setInitialState(initial);
			domain.setCurrentState(initial);
			HubConsolidation.setLocalSearchMode();
			return domain;

		} else if (domainString.equals("11")) {
			String file = "E:\\Programmierung\\Java\\Projekte\\AI\\data\\TSP\\50nodesTSP.txt";
			double best = 66.105812;
			return TSPProblem.getInstance(file, best, randomSource);

		} else if (domainString.equals("12")) {
			return FunctionProblem.getInstance(randomSource);

		} else {
			System.out.println("Invalid domain!");
			printUsage();
			System.exit(0);
		}
		return null;
	}

	private static IStrategy getStrategy(String strategyString,
		float parameter1, float parameter2, IRandomSource randomSource)
	{
		if (strategyString.equals("1")) {
			return new BreadthFirstSearch();

		} else if (strategyString.equals("2")) {
			return new BestFirstSearch();

		} else if (strategyString.equals("3")) {
			return new DepthFirstSearch();

		} else if (strategyString.equals("4")) {
			return new DepthLimitedSearch((int) parameter1);

		} else if (strategyString.equals("5")) {
			return new IterativeDeepeningSearch();

		} else if (strategyString.equals("10")) {
			return new AStarSearch();

		} else if (strategyString.equals("11")) {
			return new AStarSearchSpaceOptimized();

		} else if (strategyString.equals("12")) {
			return new GreedySearch();

		} else if (strategyString.equals("19")) {
			return new AStarSearchWork();

		} else if (strategyString.equals("20")) {
			return new HillClimbing(-1);

		} else if (strategyString.equals("21")) {
			// do random operator selection
			return new HillClimbingRandomRestart(100, -1);

		} else if (strategyString.equals("22")) {
			// do random operator selection
			return new HillClimbingRandom(1000000, 60000);

		} else if (strategyString.equals("23")) {
			// do random walk
			return new RandomWalk(30, 100000, 60000);

		} else if (strategyString.equals("27")) {
			return new SimulatedAnnealing(new LinearTemperatureReduction(
				parameter1, 0.0), (int) parameter2, -1);

		} else if (strategyString.equals("28")) {
			return new TabuSearch((int) parameter1, (int) parameter2, 60000);

		} else if (strategyString.equals("30")) {
			// do basic variable neighborhood search
			return new BasicVariableNeighborhoodSearch(new HillClimbing(-1), 175,
				60000);

		} else if (strategyString.equals("40")) {
			// do basic variable neighborhood search
			GeneticSearchParameter params = new GeneticSearchParameter(
				randomSource);
			params
				.setSelectionStrategy(new MonteCarloSelection(true, randomSource));
			params.setReproductionStrategy(new IntMultiCrossover(randomSource));
			params.setMutationStrategy(new RandomMutation(params
				.getIndividuumMutationProbability(), params
				.getGeneMutationProbability(), randomSource));
			return new GeneticSearch(params);

		} else if (strategyString.equals("50")) {
			return new MiniMaxSearch();

		} else if (strategyString.equals("51")) {
			return new AlphaBetaSearch();

		} else {
			System.out.println("Invalid strategy!");
			printUsage();
			System.exit(0);
		}
		return null;
	}

	/**
	 * The main method to start the search or two player game selected
	 * @param args the list of command line options
	 */
	private static void printUsage()
	{
		System.out
			.println("Usage: SearchMain <domain> [<size>] [<strategy> [<param>]] where");
		System.out.println("values for <domain>:");
		System.out.println("	 1: VacuumWorld with <size> (mandatory) rooms ");
		System.out.println("	 2: Puzzle of 8 ");
		System.out.println("	10: Logistik");
		System.out.println("	11: TSP");
		System.out.println("	12: Function approximation");
		System.out.println("	20: TicTacToe");
		System.out.println("	21: Vier Gewinnt <playlevel> (mandatory)");
		System.out
			.println("	21: VierGewinnt with <level> [0-2] computer player strength");
		System.out.println("values for <strategy>:");
		System.out.println("	 1: BreadthFirst");
		System.out.println("	 2: BestFirst");
		System.out.println("	 3: DepthFirst");
		System.out.println("	 4: DepthLimited <param>=depth limit");
		System.out.println("	 5: Iterative Deepening");
		System.out.println("	10: A*");
		System.out.println("	11: A* space optimized");
		System.out.println("	12: Greedy Search");
		System.out.println("	20: Hill-Climbing");
		System.out.println("	21: Hill-Climbing with random restart");
		System.out.println("	22: Hill-Climbing with random operator selection");
		System.out.println("	23: Random walk");
		System.out
			.println("	27: Simulated Annealing <temp>=initial temperature <it>=number of iterations");
		System.out
			.println("	28: Tabu Search <size>=size of tabu list <it>=number of iterations");
		System.out.println("	30: Variable Neighborhood Search");
		System.out.println("	40: Genetic Search");
		System.out.println("	50: MiniMax");
		System.out.println("	51: AlphaBeta");
	}
}
